# Copyright 2017 The TensorFlow Authors All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Global parameters for the VGGish model.

See vggish_slim.py for more information.
"""

# Architectural constants.
NUM_FRAMES = 96  # Frames in input mel-spectrogram patch.
NUM_BANDS = 64  # Frequency bands in input mel-spectrogram patch.
EMBEDDING_SIZE = 128  # Size of embedding layer.

# Hyperparameters used in feature and example generation.
SAMPLE_RATE = 16000
STFT_WINDOW_LENGTH_SECONDS = 0.025
STFT_HOP_LENGTH_SECONDS = 0.010
NUM_MEL_BINS = NUM_BANDS
MEL_MIN_HZ = 125
MEL_MAX_HZ = 7500
LOG_OFFSET = 0.01  # Offset used for stabilized log of input mel-spectrogram.
EXAMPLE_WINDOW_SECONDS = 0.96  # Each example contains 96 10ms frames
EXAMPLE_HOP_SECONDS = 0.96     # with zero overlap.

# Parameters used for embedding postprocessing.
PCA_EIGEN_VECTORS_NAME = 'pca_eigen_vectors'
PCA_MEANS_NAME = 'pca_means'
QUANTIZE_MIN_VAL = -2.0
QUANTIZE_MAX_VAL = +2.0

# Hyperparameters used in training.
INIT_STDDEV = 0.01  # Standard deviation used to initialize weights.
LEARNING_RATE = 1e-4  # Learning rate for the Adam optimizer.
ADAM_EPSILON = 1e-8  # Epsilon for the Adam optimizer.

# Names of ops, tensors, and features.
INPUT_OP_NAME = 'vggish/input_features'
INPUT_TENSOR_NAME = INPUT_OP_NAME + ':0'
OUTPUT_OP_NAME = 'vggish/embedding'
OUTPUT_TENSOR_NAME = OUTPUT_OP_NAME + ':0'
AUDIO_EMBEDDING_FEATURE_NAME = 'audio_embedding'



from pathlib import PureWindowsPath, PurePosixPath
import os

def map_path_to_os(path: str) -> str:
    """Maps the given folder path to the appropriate format for the current operating system.
    :param folder_path: The folder path to map.
    :return: The mapped folder path.
    :rtype: str
    """
    
    # Check inptu validity
    if path is None:
        raise ValueError("The folder_path cannot be None.")
    if not isinstance(path, str):
        raise TypeError(f"The folder_path was expected to be a string but is {type(path)}.")
    
    # Map path to operating system
    if os.name == 'nt': # Windows
        path = str(PureWindowsPath(path))
    elif os.name == 'posix':
        path = str(PurePosixPath(path))
        
    # Output
    return path





import torch

class SimpleDenseTransformation(torch.nn.Module):

    def __init__(self, input_dimensionality, output_dimensionality):
        super().__init__()

        self._linear = torch.nn.Linear(input_dimensionality, output_dimensionality)
        self._sigmoid = torch.nn.Sigmoid()

    def forward(self, x):
        return self._sigmoid(self._linear(x))
    





import threading
import time
import sys
from typing import List, Any
import sys
import os
sys.path.append(os.path.abspath(""))
from briann.python.utilities import callbacks as bpuc

class Observer():

    def __init__(self, on_separate_thread: bool = False):
        self._listeners = {}
        
        if on_separate_thread: 
            self._original_trace = threading.gettrace()
            threading.settrace(self.trace_function)
        else: 
            self._original_trace = sys.gettrace()
            sys.settrace(self.trace_function)

    def trace_function(self, frame, event, arg, indent=[0]):
        if self._original_trace: self._original_trace(frame, event, arg)

        # Ensure method is on watch list
        identifier = f"{frame.f_code.co_name}({list(frame.f_locals.keys())})"
        
        if identifier in self._listeners.keys():
                
                entries = self._listeners[identifier]
                for entry in entries:

                    # Ensure the function is on exit
                    if event == entry.event:

                        # No argument value constraints
                        if entry.required_argument_values == None:
                            entry.callback(frame.f_locals)

                        # Argument value constraints
                        else:
                            if all([value == frame.f_locals[key] for key, value in entry.required_argument_values.items()]):
                                entry.callback(frame.f_locals)
                    
        return self.trace_function

    def link(self, method_name: str, argument_names: List[str], callback: callable, event, required_argument_values = None):
        
        target = Observer.Entry(method_name=method_name, argument_names=argument_names, callback = callback, event=event, required_argument_values=required_argument_values)

        # Ensure function name is listened to
        if target.identifier not in self._listeners.keys():
            self._listeners[target.identifier] = []
        
        # Add outlet property to function name
        self._listeners[target.identifier].append(target)

    class Entry():

        def __init__(self, method_name, argument_names, callback, event, required_argument_values):
            self.method_name = method_name
            self.argument_names = argument_names
            self.callback = callback
            self.event = event
            self.required_argument_values = required_argument_values
            
            self.identifier = f"{self.method_name}({(self.argument_names)})"

if __name__ == "1__main__":   
    class A():

        def __init__(self, x):
            self._x = x

        @property
        def x(self):
            return self._x
        
        @x.setter
        def x(self, new_value):
            print("x")
            self._x = new_value

        def add(self, num):
            print("add")
            return self.x + num

    def other_function(num):
        print("other_function")

    # Callbacks
    def set_A_x_callback(args):
        print("set_A_x_callback")

    def A_add_callback(args):
        print("A_add_callback")

    def other_callback(args):
        print("Other callback")

    oldtrace = None
    try:
        import pydevd
        debugger=pydevd.GetGlobalDebugger()
        if debugger is not None:
            oldtrace = [debugger.trace_dispatch]

    except ImportError:
        pass

    if oldtrace is None:
        oldtrace = [frame.f_trace]

    # Regular example
    a = A(x=3)
    o = Observer()
    o.link(method_name='x', argument_names=['self', 'new_value'], callback=set_A_x_callback, event='return', required_argument_values={'self':a})
    o.link(method_name='add', argument_names=['self', 'num'], callback=A_add_callback, event='return', required_argument_values={'self':a})
    o.link(method_name='other_function', argument_names=['num'], callback=other_callback, event='return')
    
    a.x = 2
    a.add(11)
    other_function(num=5)

    print("\n\n\nMulti-threading example")
    # Multi-threading example
    a = A(x=3)
    t1 = threading.Thread(target=lambda x, a=a: setattr(a, 'x', x), args=[10])
    t2 = threading.Thread(target=A.add, args=[a,10])
    t3 = threading.Thread(target=other_function, args=[4])

    o = Observer(on_separate_thread=True)
    o.link(method_name='x', argument_names=['self', 'new_value'], callback=set_A_x_callback, event='return', required_argument_values={'self':a})
    o.link(method_name='add', argument_names=['self', 'num'], callback=A_add_callback, event='return', required_argument_values={'self':a})
    o.link(method_name='other_function', argument_names=['num'], callback=other_callback, event='return')
    
    t1.start()
    t2.start()
    t3.start()

    t1.join()
    t2.join()
    t3.join()

    print("Done!")

class A():

    def __init__(self, x):
        self.x=x

    @property
    def x(self): return self._x 

    @x.setter
    def x(self, new_value): 
        print("inside a.x")
        self._x = new_value

class B(A):

    def __init__(self, x):
        super().__init__(x=x)
        bpuc.CallbackManager.add_callback_to_attribute(B, self, 'x', B.callback)

    def callback(obj, name, value):
        print(obj, name, value)

if __name__ == "2__main__":
    a = A(x=2)
    b = B(x=1)

    a.x=5
    b.x = 3

    import torch

    linear = torch.nn.Linear(3,5, bias=False)
    for p in linear.parameters():
        print(p.shape)
    print(linear.bias==None)
    
import json

class One():
    def __init__(self,x):
        self.x=x

class Two():
    def __init__(self, one):
        self.one=one

if __name__ == "__main__":
    one = One(4)
    two = Two(one)

    json.dumps(two)


import sys, os
sys.path.append(os.path.abspath(""))
from briann.utilities import core as bpuc
import torch
from typing import Tuple, List, Dict

class Splitter(bpuc.Adapter):
    """This class is to be used in the :py:meth:`~.Connection.forward` method. It takes an input x that is the current :py:meth:`.TimeFrame.state` from the sending 
    :py:class:`.Area.output_time_frame_accumulator` and then splits off the part that is relevant to the calling :py:class:`.Connection` object."""
    
    def __init__(self) -> "Splitter":
        super().__init__()
        
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        pass

class IndexBasedSplitter(torch.nn.Module):
    """The IndexBasedSplitter maps its input to its output by using indices in the :py:meth:`~.IndexBasedSplitter.forward`. To configure this mapping,
    the following arguments are used.

    :param input_flatten_axes: Sets the :py:meth:`~.IndexBasedSplitter.input_flatten_axes` of this object.
    :type input_flatten_axes: Tuple[int,int]
    :param input_indices: Sets the :py:meth:`~.IndexBasedSplitter.input_indices` of this object.
    :type input_indices: List[int]
    :param output_flatten_axes: Sets the :py:meth:`~.IndexBasedSplitter.output_flatten_axes` of this object.
    :type output_flatten_axes: Tuple[int,int]
    :param output_shape: Sets the :py:meth:`~.IndexBasedSplitter.output_shape` of this object.
    :type output_shape: List[int]
    """
    
    def __init__(self, input_flatten_axes: Tuple[int,int],
                 input_indices: List[int],
                 output_flatten_axes: Tuple[int,int],
                 output_shape: List[int]) -> "IndexBasedSplitter":
        
        # Super
        super().__init__()
        
        # Properties
        self.input_flatten_axes = input_flatten_axes
        self.input_indices = input_indices
        self.output_flatten_axes = output_flatten_axes
        self.output_shape = output_shape

    @property
    def input_flatten_axes(self) -> Tuple[int,int]:
        """:return: The axes along which the input will be flattened inside :py:meth:`~.IndexBasedSplitter.forward` before selecting entries from it. This is a Tuple of two ints, where the first int is the axis at which flattening starts and the second int is the axis (inclusive) at which it will end. 
        :rtype: Tuple[int, int]"""

        return self._input_flatten_axes
    
    @input_flatten_axes.setter
    def input_flatten_axes(self, new_value: Tuple[int,int]) -> None:
        # Input validity
        # - Type
        if not isinstance(new_value, Tuple) or not len(new_value) == 2 or not all([isinstance(entry, int) for entry in new_value]): raise TypeError(f"The input_flatten_axes of a IndexBasedSplitter are expected to be of type Tuple[int,int] but were {new_value}.")

        # - Value
        if not all([entry >= 1 for entry in new_value]): raise ValueError(f"When setting the input_flatten_axes, no value below 1 is permitted, yet {new_value} was provided.")

        # Set property
        self._input_flatten_axes = new_value

    @property
    def input_indices(self) -> Dict[int, List[int]]:
        """:return: A list of indices that specify which entries of the flattened input :py:class:`torch.Tensor` shall be moved to in the flattened output tensor. The length of this list must factor into :py:meth:`~IndexBasedSplitter.output_shape` along the :py:meth:`~IndexBasedSplitter.output_flatten_axes`.
        :rtype: List[int]"""
        return self._input_indices
    
    @input_indices.setter
    def input_indices(self, new_value: List[int]) -> None:

        # Input validity
        # - Type
        if not all([isinstance(entry, int) for entry in new_value]): raise TypeError(f"The output_indices of an IndexBasedMerger was expected to be a List[int], but was tried to be set to {new_value}")

        # - Values
        for tensor_index in new_value:
            if tensor_index < 0: raise ValueError(f"An illegal index ({tensor_index}) was tried to be set as tensor_index when setting the output_indices of an IndexBasedSplitter.")

        # Set value
        self._input_indices = new_value

    @property
    def output_flatten_axes(self) -> Tuple[int,int]:
        """:return: A Tuple of two axes, namely the start_axis and end_axis (inclusive). Axes spanned by the start_axis and end_axis are the axes along which the output tensor will be flattened before its entries are filled with the corresponing ones from the input tensor. The calculation of these axes DOES include the initial batch axis and is thus assumed to lead to axes greater than 0.
        :rtype: Tuple[int,int]"""
        return self._output_flatten_axes
    
    @output_flatten_axes.setter
    def output_flatten_axes(self, new_value: Tuple[int,int]) -> None:

        # Input validity
        # - Type
        if not len(new_value) == 2 or not all([isinstance(entry, int) for entry in new_value]): raise TypeError(f"The output_flatten_axes of an IndexBasedSplitter was expected to be a Tuple[int,int], but was tried to be set to {new_value}.")

        # - Values
        for axis in new_value:
            if axis < 1: raise ValueError(f"An illegal axis ({axis}) was tried to be set as input_flatten_axes of an IndexBasedSplitter.")

        # Set value
        self._output_flatten_axes = new_value

    @property
    def output_shape(self) -> List[int]:
        """:return: The desired shape for the :py:class:`torch.Tensor` that the :py:meth:`~.IndexBasedSplitter.forward` method shall output.
        :rtype: List[int]"""
        return self._output_shape
    
    @output_shape.setter
    def output_shape(self, new_value: List[int]) -> None:

        # Input validity
        # - Type
        if not all([isinstance(entry, int) for entry in new_value]): raise TypeError(f"The output_shape of an IndexBasedSplitter was expected to be a List[int], but was tried to be set to {new_value}.")

        # - Values
        for dimensionality in new_value:
            if dimensionality < 0: raise ValueError(f"An illegal shape ({new_value}) was tried to be set for an IndexBasedSplitter.")

        # Set value
        self._output_shape = new_value

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """This method first flattens the input tensor `x` along :py:meth:`~.IndexBasedSplitter.input_flatten_axes`, then extracts the corresponding entries using 
        the indices from :py:meth:`~.IndexBasedSplitter.output_indices` and then unflattens the output along the axes specified in 
        :py:meth:`~.IndexBasedSplitter.output_flatten_axes` to arrive at the output shape specified in :py:meth:`~.IndexBasedSplitter.output_shape`."""

        # Flatten x
        x_axis = self.input_flatten_axes[0]
        x = torch.flatten(input=x, start_dim=x_axis, end_dim=self.input_flatten_axes[1])

        # Split by extracting relevant section
        indices = self.input_indices
        y = torch.index_select(input=x, dim=x_axis, index=torch.tensor(indices))

        # Unflatten y
        start_axis = self.output_flatten_axes[0]
        end_axis = self.output_flatten_axes[1] + 1 # +1 since the end_index should be inclusive
        shape_along_unflattened_axes = self.output_shape[start_axis-1:end_axis-1] # -1 since the provided output shape does not include the initial batch_size axis
        y = torch.unflatten(input=y, dim=start_axis, sizes=shape_along_unflattened_axes)

        # Output
        return y



# https://edwin-de-jong.github.io/blog/mnist-sequence-data/
import sys
import os
sys.path.append(os.path.abspath(""))
from briann.utilities import random as bpur, vggish_input
from typing import Tuple
import torch
from torch.utils.data import Dataset, DataLoader
from typing import List
from src.briann.utilities import file_management as bpufm
import numpy as np

def collate_function(sequences: List[Tuple[torch.Tensor, torch.Tensor]], **kwargs) -> Tuple[torch.Tensor, torch.Tensor]:
    """Collate function to pad sequences to the same length.
    
    :param sequences: The sequences to collate.
    :type sequences: List[Tuple[torch.Tensor, torch.Tensor]]
    :return: The collated sequences.
    :rtype: Tuple[torch.Tensor, torch.Tensor]
    """
    # Unzip sequences
    X, y = zip(*sequences)

    # Pad X
    X = torch.nn.utils.rnn.pad_sequence(sequences=X, **kwargs)

    # Stack y
    y = torch.stack(y)

    return X, y

class Sinusoids(Dataset):
    """A dataset of sinusoids with varying frequencies. Each instance is a time-series of a sinusoid with one frequency and added noise.
    The frequency is determined by the index of the instance. The dataset can return the input data, the labels, or both.
    
    :param instance_count: Sets the :py:meth:`~.Sinusoids.instance_count` property.
    :type instance_count: int
    :param duration: Sets the :py:meth:`~.Sinusoids.duration` property.
    :type duration: float
    :param sampling_rate: Sets the :py:meth:`~.Sinusoids.sampling_rate` property.
    :type sampling_rate: float
    :param frequency_count: Sets the :py:meth:`~.Sinusoids.frequency_count` property.
    :type frequency_count: int
    :param noise_range: Sets the :py:meth:`~.Sinusoids.noise_range` property.
    :type noise_range: float
    :param return_X: Sets the :py:meth:`~.Sinusoids.return_X` property.
    :type return_X: bool
    :param return_y: Sets the :py:meth:`~.Sinusoids.return_y` property.
    :type return_y: bool

    :return: The sinusoid dataset.
    :rtype: Sinusoids
    """

    def __init__(self, instance_count: int, 
                 duration: float, 
                 sampling_rate: float, 
                 frequency_count: int, 
                 noise_range: float, 
                 return_X: bool = True,
                 return_y: bool = True,
                 ) -> None:
        # Call super
        super(Sinusoids, self).__init__()

        # Check input validity
        if not isinstance(instance_count, int): raise TypeError(f"The instance_count was expected to be an integer but is {type(instance_count)}.")
        if instance_count <= 0: raise ValueError(f"The instance_count was expected to be a positive integer but is {instance_count}.")

        if not isinstance(duration, (int, float)): raise TypeError(f"The duration was expected to be a number but is {type(duration)}.")
        if duration <= 0: raise ValueError(f"The duration was expected to be a positive number but is {duration}.")

        if not isinstance(sampling_rate, (int, float)): raise TypeError(f"The sampling_rate was expected to be a number but is {type(sampling_rate)}.")
        if sampling_rate <= 0: raise ValueError(f"The sampling_rate was expected to be a positive number but is {sampling_rate}.")

        if not isinstance(frequency_count, int): raise TypeError(f"The frequency_count was expected to be an integer but is {type(frequency_count)}.")
        if frequency_count <= 0: raise ValueError(f"The frequency_count was expected to be a positive integer but is {frequency_count}.")

        if not isinstance(noise_range, (int, float)): raise TypeError(f"The noise_range was expected to be a number but is {type(noise_range)}.")
        if noise_range < 0: raise ValueError(f"The noise_range was expected to be a non-negative number but is {noise_range}.")

        if not isinstance(return_X, bool): raise TypeError(f"The return_X was expected to be a boolean but is {type(new_value)}.")
        
        if not isinstance(return_y, bool): raise TypeError(f"The return_y was expected to be a boolean but is {type(new_value)}.")
        
        # Set properties
        self._instance_count = instance_count
        self._duration = (float)(duration)
        self._sampling_rate = sampling_rate
        self._frequency_count = frequency_count
        self._noise_range = noise_range
        self._return_X = return_X
        self._return_y = return_y

    @property
    def instance_count(self) -> int:
        return self._instance_count
    
    @property
    def duration(self) -> int:
        return self._duration
    
    @property
    def sampling_rate(self) -> int:
        return self._sampling_rate
    
    @property
    def frequency_count(self) -> int:
        return self._frequency_count
    
    @property
    def noise_range(self) -> float:
        return self._noise_range

    @property
    def return_X(self) -> bool:
        return self._return_X
    
    @property
    def return_y(self) -> bool:
        return self._return_y
    
    def __len__(self):
        return self.instance_count
    
    def __getitem__(self, idx):
        # Choose frequency
        frequency = idx % self._frequency_count + 1
        y = frequency

        # Generate sinusoid
        X = torch.zeros([(int)(self._sampling_rate * self._duration), 1])
        ts = torch.linspace(start=0.0, end=self._duration, steps=(int)(self._sampling_rate * self._duration))
        X[:,0] = torch.sin(ts * 2 * torch.pi * frequency)
        noise = torch.Tensor(bpur.StatelessRandom.pseudo_uniform(lower_bound=-self._noise_range/2, upper_bound=self._noise_range/2, seed=idx, count=X.shape[0]))
        X[:,0] += noise

        # Output
        if self.return_X and not self.return_y:
            return X
        if not self.return_X and self.return_y:
            return y
        if self.return_X and self.return_y:
            return X, y

class VGGishSpectrograms(Dataset):
    """A dataset of spectrogram slices as expected by VGGish, each instance is a spectrogram taken from the waveforms of the given `sinusoids` dataset. 
    The `waveform` is treated as sampled at 16 kHz.

    :param sinusoids: A dataset of sinusoids to convert to spectrograms.
    :type sinusoids: Dataset
    :return: The spectrogram dataset.
    :rtype: Sinusoids
    """

    def __init__(self, sinusoids: Dataset
                 ) -> None:
        # Call super
        super(VGGishSpectrograms, self).__init__()

        # Check input validity
        if not isinstance(sinusoids, Sinusoids): raise TypeError(f"The sinusoids argument was expected to be a Sinusoids Dataset instance but is of type {type(sinusoids)}.")
        self._sinusoids_ = sinusoids
        
    @property
    def return_X(self) -> bool:
        return self._sinusoids_._return_X
    
    @property
    def return_y(self) -> bool:
        return self._sinusoids_._return_y
    
    def __len__(self):
        return len(self._sinusoids_)
    
    def __getitem__(self, idx):
        # Draw sinusoid from dataset
        if self.return_X and self.return_y:
            waveform, y = self._sinusoids_[idx]
            X = vggish_input.waveform_to_examples(waveform.numpy())
            X = torch.Tensor(X)
            return X, y
        elif self.return_X and not self.return_y:
            waveform = self._sinusoids_[idx]
            X = vggish_input.waveform_to_examples(waveform.numpy(), sample_rate=16000)
            X = torch.Tensor(X)
            return X
        elif not self.return_X and self.return_y:
            y = self._sinusoids_[idx]
            return y

class PenStrokeMNIST(Dataset):
    
    def __init__(self, folder_path: str) -> "PenStrokeMNIST":
        # Call super
        super().__init__()

        # Set attributes
        self.folder_path = folder_path
        
    @property
    def folder_path(self) -> str:
        """The path to the folder where the data is stored.

        :return: The path to the folder where the data is stored.
        :rtype: str
        """
        return self._folder_path

    @folder_path.setter
    def folder_path(self, new_value: str) -> None:
        # Input validity
        if not isinstance(new_value, str): raise TypeError(f"The folder_path was expected to be a string but is {type(new_value)}.")
        if not os.path.isdir(new_value): raise ValueError(f"The path {new_value} is not a directory.")
        
        # Set property
        self._folder_path = new_value

    def __len__(self) -> int:
        """The number of instances in the dataset.
        :return: The number of instances in the dataset.
        :rtype: int
        """

        # Get all file names in the data folder
        file_names = os.listdir(self.folder_path)

        # Filter for files that start with 'trainimg-' or 'testimg-'
        instance_count = 0
        for file_name in file_names:
            if 'inputdata' in file_name: instance_count += 1
        
        # Output
        return instance_count
    
    def __getitem__(self, index: int) -> Tuple[torch.Tensor, torch.Tensor]:

        file_path = os.path.join(self.folder_path, f"img-{index}-targetdata.txt" )
        
        with open(file=file_path, mode='r') as file_handle:
            text_data = file_handle.read()

        # Parse
        # The first 10 columns one-hot encode the label. All but the first rows can be ignored.
        # The next 4 columns encode the dx, dy, end of stroke (0,1) and end of sequence (0, 1). The end of sequence can be ignored
        lines = text_data.split('\n')
        y = torch.Tensor([(float)(entry) for entry in lines[0].split(' ')[:10]])
        
        x = [None] * len(lines)
        for l, line in enumerate(lines):
            if len(line) > 0: # Exclude possible empty lines
                x[l] = [(float)(entry) for entry in line.split(' ')[10:13]]
        del x[l:]
        
        x = torch.Tensor(x)

        return x,y

if __name__=="__main__":
    path = bpufm.map_path_to_os(path=os.path.join("tests","data","Pen Stroke MNIST"))
    dataset = PenStrokeMNIST(folder_path=path)
    #dataset = Sinusoids(instance_count=10, duration=3.0, sampling_rate=50, frequency_count=2, noise_range=0.02)
    data_loader = DataLoader(dataset, batch_size=3, collate_fn=lambda sequences: collate_function(sequences=sequences, batch_first=True))
    
    for X, y in data_loader:
        print(X.shape, y.shape)


from typing import Any


class CallbackManager():
    """This class allows to add callbacks to attribute setters and method calls."""

    _class_to_regular_setattr = {}
    _obj_to_attribute_name_to_callbacks = {}

    def __setattr__with_callbacks(target_class, obj, name, value) -> None:
        """Sets an attribute and calls any callbacks associated with it.
        
        :param target_class: The class of the instance.
        :type target_class: Any
        :param obj: The instance to set the attribute on.
        :type obj: Any
        :param name: The name of the attribute to set.
        :type name: str
        :param value: The value to set the attribute to.
        :type value: Any
        :rtype: None"""

        # Call regular setattr
        CallbackManager._class_to_regular_setattr[target_class](obj, name, value)

        # If object is inspected
        if obj in CallbackManager._obj_to_attribute_name_to_callbacks.keys():

            # If attribute is inspected
            if name in CallbackManager._obj_to_attribute_name_to_callbacks[obj].keys():

                # Call the callbacks
                for callback in CallbackManager._obj_to_attribute_name_to_callbacks[obj][name]:
                    callback(obj, name, value)

        
    def add_callback_to_attribute(target_class: Any, target_instance: Any, attribute_name: str, callback: callable) -> None:
        """Adds a callback to an attribute setter of a specific instance of a class.
        The callback must have the signature: callback(obj: Any, name: str, value: Any)
        where obj is the instance, name is the attribute name, and value is the new value. Any returned values from the callback are ignored.
        
        :param target_class: The class of the instance.
        :type target_class: Any
        :param target_instance: The instance to inspect.
        :type target_instance: Any
        :param attribute_name: The name of the attribute to inspect.
        :type attribute_name: str
        :param callback: The callback to call when the attribute is set. The callback must have the signature: callback(obj: Any, name: str, value: Any), where obj is the instance on which the attribute setter is called, name is the name of the attribute and value is the new value to be assigned to the attribute.
        :type callback: callable
        :rtype: None
        """

        # Ensure target class is inspected
        if target_class not in CallbackManager._class_to_regular_setattr.keys():
            CallbackManager._class_to_regular_setattr[target_class] = target_class.__setattr__
            target_class.__setattr__ = lambda obj, name, value, target_class=target_class: CallbackManager.__setattr__with_callbacks(target_class=target_class, obj=obj, name=name, value=value)
        
        # Ensure object is inspected
        if target_instance not in CallbackManager._obj_to_attribute_name_to_callbacks.keys(): 
            CallbackManager._obj_to_attribute_name_to_callbacks[target_instance] = {}
        
        # Ensure name is inspected
        if attribute_name not in CallbackManager._obj_to_attribute_name_to_callbacks[target_instance].keys():
            CallbackManager._obj_to_attribute_name_to_callbacks[target_instance][attribute_name] = []

        # Enter callback
        CallbackManager._obj_to_attribute_name_to_callbacks[target_instance][attribute_name].append(callback)

    def add_callback_to_method(target_instance: Any, method_name: str, callback: callable) -> None:
        """Adds a callback to a method of a specific instance of a class.
        
        :param target_instance: The instance to inspect.
        :type target_instance: Any
        :param method_name: The name of the method to inspect.
        :type method_name: str
        :param callback: The callback to call when the method is called. The callback must have the signature: callback(obj: Any, *args, **kwargs), where obj is the object in which the method is called, and args and kwargs are the positional and key-word arguments of the method, respectively. For arguments that are optional for the method, the default value assigned to the corresponding callback's argument takes priority over that asigned to the method's argument. Any returned values from the callback are ignored.
        :type callback: callable
        :rtype: None
        """
        
        # Ensure method is wrapped
        if not hasattr(target_instance, f"__regular_{method_name}"):
            global new_method, ttarget_instance
            ttarget_instance = target_instance
            def new_method(*args, **kwargs):
                # Call the method
                global ttarget_instance, aargs, kkwargs
                ttarget_instance, aargs, kkwargs = target_instance, args, kwargs
                exec(f"global ttarget_instance, aargs, kkwargs; ttarget_instance.__regular_{method_name}(*aargs, **kkwargs)")

                # Call the callbacks
                args = (target_instance,) + args
                callbacks = getattr(target_instance, f"_{method_name}_callbacks")
                for callback in callbacks: callback(*args, **kwargs)
        
            exec(f"global ttarget_instance, new_method; ttarget_instance.__regular_{method_name} = ttarget_instance.{method_name}; ttarget_instance.{method_name} = new_method")

        # Ensure target instance has callbacks for this method
        if not hasattr(target_instance, f"_{method_name}_callbacks"):
            setattr(target_instance, f"_{method_name}_callbacks", [])

        # Add callback to list
        getattr(target_instance, f"_{method_name}_callbacks").append(callback)
        


import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import (FigureCanvasTkAgg) 
import tkinter as tk
import customtkinter
import numpy as np

from briann.utilities import callbacks as bpuc
customtkinter.set_appearance_mode("Light")  # Modes: "System" (standard), "Dark", "Light"
customtkinter.set_default_color_theme("blue")  # Themes: "blue" (standard), "green", "dark-blue"
import sys, os
sys.path.append(os.path.abspath(""))
from briann.network import core as bpnc
import networkx as nx   
import tkinter as tk
from typing import Tuple, List, Iterator
from CTkMenuBar import *
import json
from briann.utilities import file_management as bpufm
import threading, time
from abc import ABC, abstractmethod


# Get the DPI
root = tk.Tk()
root.withdraw()  # Hide the root window
DPI = root.winfo_fpixels('1i')  # '1i' means 1 inch
root.destroy()  # Destroy the hidden window

class Animator(customtkinter.CTk):
    """This class can be used to animate :py:class:`~src.briann.python.components.BrIANN`s. The animator displays a graph for the different :py:class:`~src.briann.python.components.Area`s and :py:class:`~src.briann.python.components.Connections`s.
    The layout of the graph can be adjusted with mouse-clicks inside the animator. It is also possible to add visualisations for area states by right-clicking the areas. 
    When stepping through the simulation, the animator indicates with a color-code which area is currently processing its inputs and updates to their states are displayed in the corresponding visualizers.
    
    :param briann: The briann instance to be animated.
    :type briann: :py:class:`~src.briann.python.components.BrIANN`
    :param data_iterator: An iterator that provides the data (X, y (ignored)) to be streamed to the briann model during the simulation.
    :type data_iterator: Iterator
    :return: An instance of this class.
    :rtype: :py:class:`.Animator`
    """

    @property
    def briann(self):
        """:return: The briann instance that is being animated.
        :rtype: :py:class:`~.Animator`"""

        return self._briann

    def __init__(self, briann: bpnc.BrIANN, data_iterator: Iterator) -> "Animator":
        # Call super
        super().__init__()
        
        # Start simulation
        self._data_iterator = data_iterator
        X, y = next(data_iterator)
        briann.load_next_stimulus_batch(X=X)

        # Configure window
        self.title("BrIANN Animator")
        self.geometry(f"{(int)(18*DPI)}x{(int)(10*DPI)}")

        # Canvas
        canvas = Canvas(master=self, xscrollincrement=1, yscrollincrement=1)
        canvas.pack(expand=True, fill='both')
        self._briann = briann
        
        # Network Visualizer
        network_visualizer = NetworkVisualizer(briann=briann, canvas=canvas, initial_x=0.0, initial_y=0.0, width=4, height=4, area_size=0.5)

        # Time Label
        self._time_label = customtkinter.CTkLabel(canvas, text=f"Time: {briann.current_simulation_time:.3f} s", anchor=tk.CENTER, font=("Arial", 16))
        self._time_label.place(relx=0.5, rely=0.01, anchor=tk.N)

        # Add self as subscriber for briann simulation time
        bpuc.CallbackManager.add_callback_to_attribute(target_class=type(briann), target_instance=briann, attribute_name='_current_simulation_time', callback=self.on_current_simulation_time_update)
        
        # Controller
        controller_frame = ControllerFrame(briann=briann, network_visualizer=network_visualizer, master=self, corner_radius=0)
        controller_frame.pack(fill='x', expand=False, side=tk.BOTTOM)
        
        # Quit handler
        self.protocol(name='WM_DELETE_WINDOW', func=self._on_window_close) 
    
    def on_current_simulation_time_update(self, obj, name, value) -> None:
        
        # Set time-label
        self._time_label.configure(text=f"Time: {value:.3f} s")
        
    def _on_window_close(self) -> None:
        """Ensures the animator closes all visualizations properly before quitting the app.
        :rtype: None"""

        # Close all matplotlib figures to prevent errors
        plt.close('all')  
        self.after(50, self.destroy)

class ControllerFrame(customtkinter.CTkFrame):
    """A frame that holds the different buttons used to control the animation.
    
    :param briann: The briann instance for which the animation shall be controlled.
    :type briann: :py:class:`~.src.briann.python.components.BrIANN`
    :param network_visualizer: The visualizer that displays the current state of `briann`.
    :type network_visualizer: :py:class:`~.NetworkVisualizer`
    returns: An instance of this class.
    :rtype: :py:class:`.ControllerFrame`
    """

    def __init__(self, briann: bpnc.BrIANN, network_visualizer: "NetworkVisualizer", **kwargs) -> "ControllerFrame":
        
        # Super
        super().__init__(**kwargs)

        # Set properties
        self._briann = briann
        self._network_visualizer = network_visualizer
        self._is_playing = False

        # Create buttons
        self._play_button = customtkinter.CTkButton(self, text="Play", command=self._on_play_button_click)
        self._play_button.pack(expand=True, side=tk.LEFT, padx=0, pady=10, anchor=tk.CENTER)
        customtkinter.CTkButton(self, text="Pause", command=self._on_pause_button_click).pack(expand=True, side=tk.LEFT, padx=0, pady=10, anchor=tk.CENTER)
        customtkinter.CTkButton(self, text="Next Time-Frame", command=self.on_next_time_frame_button_click).pack(expand=True, side=tk.LEFT, padx=0, pady=10, anchor=tk.CENTER)
        customtkinter.CTkButton(self, text="Next Stimulus", command=self.on_next_stimulus_button_click).pack(expand=True, side=tk.LEFT, padx=0, pady=10, anchor=tk.CENTER)
        
    def _on_play_button_click(self) -> None:
        """Starts a loop that steps through the animation until stopped via :py:meth:`~.ControllerFrame._on_pause_button_click`.
        :rtype: None"""

        # Set property
        self._is_playing = True

        # Start loop on separate thread
        threading.Thread(target=self._play).start()

        # Disable button
        self._play_button.configure(state='disabled')
        
    def _play(self) -> None:
        """Steps through the animation for as long as :py:attr:`~Canvas._is_playing` is True.
        :rtype: None"""

        # Loop
        while self._is_playing:

            # Step
            self.on_next_time_frame_button_click()

            # Wait
            time.sleep(1)

    def _on_pause_button_click(self) -> None:
        """Halts the animation, in case it is currently running.
        :rtype: None"""

        # Set property
        self._is_playing = False

        # Enable button
        self._play_button.configure(state='normal')

    def on_next_time_frame_button_click(self) -> None:
        """Loads the next :py:class:`~src.briann.python.components.TimeFrame` into the :py:class:`~src.briann.python.components.BrIANN` model.
        :rtype: None"""

        # Step
        self._briann.step() # The area state subscribers will automatically update the UI
        
    def on_next_stimulus_button_click(self) -> None:
        """Loads the next stimulus into the :py:class:`~src.briann.python.components.BrIANN` model.
        :rtype: None"""

        # Ensure current simulation is paused
        self._is_playing = False
        
        # Load next stimulus
        X, y = next(self._data_iterator)
        self._briann.load_next_stimulus_batch(X=X)

class Canvas(tk.Canvas):
    """This class creates a canvas on which network components will be displayed. The canvas can be dragged around and has a reference grid in the background.
    
    :return: An instance of this class.
    :rtype: :py:class:`Canvas`
    """

    def __init__(self, **kwargs) -> "Canvas":
        super(Canvas, self).__init__(**kwargs)
        
        # Add reference grid
        self.create_reference_grid(x_min=-15, x_max=15, y_min=-15, y_max=15, spacing=1.0)
        
        self._drag_start_x = None; self._drag_start_y = None # For panning
        self.bind("<ButtonPress-1>", self._on_drag_start)
        self.bind("<B1-Motion>", self._on_drag_motion)
        self.bind("<ButtonRelease-1>", lambda event: setattr(self, "_drag_start_x", None) or setattr(self, "_drag_start_y", None))  # Reset mouse position on release
        self.bind(sequence="<Configure>", func= self.center_scroll_region)  # Update scroll region to fit all items
        
    def create_reference_grid(self, x_min: float, x_max: float, y_min: float, y_max: float, spacing: float = 1.0, color: str = 'lightgray') -> None:
        """Creates a reference grid on the canvas with the given parameters.
        
        :param x_min: The minimum x-coordinate of the grid in cartesian space (inches).
        :type x_min: float
        :param x_max: The maximum x-coordinate of the grid in cartesian space (inches).
        :type x_max: float
        :param y_min: The minimum y-coordinate of the grid in cartesian space (inches).
        :type y_min: float
        :param y_max: The maximum y-coordinate of the grid in cartesian space (inches).
        :type y_max: float
        :param spacing: The spacing between the grid lines in cartesian space (inches).
        :type spacing: float
        :param color: The color of the grid lines.
        :type color: str
        :rtype: None"""
        
        # Vertical lines
        x = 0
        while x <= x_max:
            x1, y1 = Canvas.cartesian_to_canvas(x=x, y=y_min)
            x2, y2 = Canvas.cartesian_to_canvas(x=x, y=y_max)
            self.create_line(x1, y1, x2, y2, fill=color, width=1)
            if x != 0:
                x1, y1 = Canvas.cartesian_to_canvas(x=-x, y=y_min)
                x2, y2 = Canvas.cartesian_to_canvas(x=-x, y=y_max)
                self.create_line(x1, y1, x2, y2, fill=color, width=1)
            x += spacing

        # Horizontal lines
        y = 0
        while y <= y_max:
            x1, y1 = Canvas.cartesian_to_canvas(x=x_min, y=y)
            x2, y2 = Canvas.cartesian_to_canvas(x=x_max, y=y)
            self.create_line(x1, y1, x2, y2, fill=color, width=1)
            if y != 0:
                x1, y1 = Canvas.cartesian_to_canvas(x=x_min, y=-y)
                x2, y2 = Canvas.cartesian_to_canvas(x=x_max, y=-y)
                self.create_line(x1, y1, x2, y2, fill=color, width=1)
            y += spacing

    def center_scroll_region(self, event: tk.Event) -> None:
        """Centers the scroll region of the canvas.

        :param event: The event that triggered the function call.
        :type event: :py:class:`tkinter.Event`
        :rtype: None"""
        
        # Get the current scroll region
        scroll_region = self.bbox(tk.ALL)
        
        # Calculate the center position
        center_x = scroll_region[0] + (scroll_region[2] - scroll_region[0]) / 2
        center_y = scroll_region[1] + (scroll_region[3] - scroll_region[1]) / 2
        
        # Set the canvas view to center the scroll region
        self.xview_scroll(-(int)(self.winfo_width()/2-center_x), "units")
        self.yview_scroll(-(int)(self.winfo_height()/2-center_y), "units")

    def _on_drag_start(self, event: tk.Event) -> None:
        """Prepares the drag operation

        :param event: The event that triggered the function call.
        :type event: :py:class:`tkinter.Event`
        :rtype: None"""

        self._press_x = event.x
        self._press_y = event.y
        self._drag_start_x = event.x
        self._drag_start_y = event.y
        
    def _on_drag_motion(self, event: tk.Event) -> None:
        """Performs a step of the drag operation.
        
        :param event: The event that triggered the function call.
        :type event: :py:class:`tkinter.Event`
        :rtype: None"""
        
        if self._drag_start_x != None and self._drag_start_y != None:
            dx = event.x - self._drag_start_x
            dy = event.y - self._drag_start_y 
            self.xview_scroll(-dx, "units")
            self.yview_scroll(-dy, "units")
            
            self._drag_start_x = event.x
            self._drag_start_y = event.y

    @staticmethod
    def cartesian_to_canvas(x: float, y: float) -> Tuple[int, int]:
        """Converts from a cartesian coordinate system to this canvas's coordinate system.
        - The cartesian coordinate system has its origin in the center of the canvas, with the x-axis pointing to the right and the y-axis pointing upwards and its units are inches.
        - The canvas coordinate system has its origin in the top left corner of the canvas, with the x-axis pointing to the right and the y-axis pointing downwards and its units are pixels.
        
        :param x: The x-coordinate in cartesian coordinates.
        :type x: float
        :param y: The y-coordinate in cartesian coordinates.
        :type y: float
        :return: The x and y coordinates in canvas space.
        :rtype: Tuple[int, int]"""

        return (int)(x*DPI), (int)(-y*DPI)

    @staticmethod
    def canvas_to_cartesian(x: int, y: int) -> Tuple[float, float]:
        """Converts from this canvas's coordinate system to a cartesian coordinate system.
        - The cartesian coordinate system has its origin in the center of the canvas, with the x-axis pointing to the right and the y-axis pointing upwards and its units are inches.
        - The canvas coordinate system has its origin in the top left corner of the canvas, with the x-axis pointing to the right and the y-axis pointing downwards and its units are pixels.
        
        :param x: The x-coordinate in canvas coordinates.
        :type x: int
        :param y: The y-coordinate in canvas coordinates.
        :type y: int
        :return: The x and y coordinates in cartesian space.
        :rtype: Tuple[float, float]"""

        return (float)(x)/DPI, -(float)(y)/DPI

class DraggableWidget():
    """Creates a visualizer that visualizes data in a rectangle whose center is at (`x`,`y`).
        The visualizer will be drawn on top of all previously drawn elements on the `canvas` and it can be dragged around with the mouse.

        :param canvas: The canvas to draw the visualizer on.
        :type canvas: Canvas
        :param widget: The widget to be placed on the canvas.
        :type widget: py:class:`tkinter.Widget`
        :param x: The x-coordinate of the visualizer's center in Cartesian space (inches).
        :type x: float, optional, defaults to 0
        :param y: The y-coordinate of the visualizer's center in Cartesian space (inches).
        :type y: float, optional, defaults to 0
        
        :return: The visualizer that was created.
        :rtype: :py:class:`.Visualizer`"""
        
    def __init__(self, canvas: Canvas, widget: tk.Widget, x: float = 0, y: float = 0) -> "DraggableWidget":
        
        # Set properties
        self._canvas = canvas
        self._initial_x, self._initial_y = Canvas.cartesian_to_canvas(x=x, y=y) # Convert to canvas space
        self._x, self._y = self._initial_x, self._initial_y # Current position in canvas space
        self._location_subscribers = []
                
        # Create window on canvas
        self._window_index = self._canvas.create_window(self._x, self._y, window=widget, anchor=tk.CENTER)
        
        # Position along Z-Axis
        widget.tk.call('lower', widget._w, None)
        widget.tk.call('lower', self._canvas._w, None)

        # Add drag listeners
        widget.bind("<ButtonPress-1>", self._on_drag_start)
        widget.bind("<B1-Motion>", self._on_drag_motion)
        widget.bind("<ButtonRelease-1>", self._on_drag_end)  # Reset mouse position on release
        
    @property
    def x(self) -> float:
        """
        :return: The x-position of the center of self in cartesian space (inches).
        :rtype: float"""
        
        # Convert to cartesian space
        x, _ = self._canvas.canvas_to_cartesian(x=self._x, y=self._y)
        
        # Output
        return x
    
    @property
    def y(self) -> float:
        """
        :return: The y-position of the center of self in cartesian space (inches)f.
        :rtype: float"""
        
        # Convert to cartesian space
        _, y = self._canvas.canvas_to_cartesian(x=self._x, y=self._y)
        
        # Output
        return y

    def _on_drag_start(self, event: tk.Event) -> None:
        """Begins a drag operation.
        
        :param event: The event that triggered the function call.
        :type event: :py:class:`tkinter.Event`
        :rtype: None"""
        
        # Save mouse position in canvas space
        self._mouse_x = self._initial_x + event.x
        self._mouse_y = self._initial_y + event.y
            
    def add_location_subscriber(self, subscriber: "DraggableWidget.LocationSubscriber") -> None:
        """Adds a subscriber that will be notified once the location of this DraggableWidget changes.

        :param subscriber: The subscriber to be added to the subscription list.
        :type subscriber: :py:class:`.DraggableWidget.LocationSubscriber`
        :rtype: None"""

        self._location_subscribers.append(subscriber)

    def _on_drag_motion(self, event: tk.Event) -> None:
        """Begins a drag operation.
        
        :param event: The event that triggered the function call.
        :type event: :py:class:`tkinter.Event`
        :rtype: None"""
        
        # Compute delta
        dx = self._initial_x + event.x - self._mouse_x
        dy = self._initial_y + event.y - self._mouse_y
        
        # Move self on canvas
        self._canvas.move(self._window_index, dx, dy)

        # Update position of self
        self._x += dx
        self._y += dy

        # Update subscribers
        for subscriber in self._location_subscribers:
            subscriber.on_location_update(draggable_widget=self)
        
    def _on_drag_end(self, event: tk.Event) -> None:
        """Ends a drag operation.
        
        :param event: The event that triggered the function call.
        :type event: :py:class:`tkinter.Event`
        :rtype: None"""
        
        # Reset mouse position
        self._mouse_x = 0
        self._mouse_y = 0
        
        # Update initial position
        self._initial_x, self._initial_y = self._x, self._y

    class LocationSubscriber(ABC):
        """An abstract base class for subscribers that want to receive the location of a :py:class:`.DraggableWidget` every time it is updated. Subscribers must implement the :py:meth:`~.DraggableWidget.LocationSubscriber.on_location_update` method.
        """

        @abstractmethod
        def on_location_update(self, draggable_widget: "DraggableWidget") -> None:
            """This method will be called every time the :py:class:`.DraggableWidget`'s :py:meth:`~.DraggableWidget.x` or :py:meth:`~.DraggableWidget.y` are updated.

            :param draggable_widget: The draggable widget that was updated.
            :type draggable_widget: :py:class:`~.DraggableWidget`
            :rtype: None
            """
            pass

class Area(DraggableWidget):
    """A visual representation of a :py:class:`~.src.briann.python.components.BrIANN` that can be placed onto the :py:class:`.Canvas`.
    It is represented by a button that can be moved around. The button changes color to indicate whether the corresponding BrIANN area is currently being updated.
    When right-clicking the button, an option menu unfolds that allows to add a visualizer for the state of the area.
    
    :param bpnc_area: The BrIANN area to be visualized.
    :type bpnc_area: :py:class:`~.src.briann.python.components.Area`
    :param canvas: The canvas on which the area shall be visualized.
    :type canvas: :py:class:'.Canvas`
    :param briann: The BrIANN that is being simulated.
    :type briann: :py:class:`~src.briann.python.components.BrIANN`
    :param x: The x-coordinate of the visualizer's center in Cartesian space (inches).
    :type x: float, optional, defaults to 0
    :param y: The y-coordinate of the visualizer's center in Cartesian space (inches).
    :type y: float, optional, defaults to 0
    :return: An instance of this class.
    :rtype: py:class:`.Area`
    """

    def __init__(self, bpnc_area: bpnc.Area, canvas: Canvas, briann: bpnc.BrIANN, x: float, y: float, size: float) -> "Area":
       
        # Set properties
        self._bpnc_area = bpnc_area
        self._briann = briann
        self._size = size

        # Link callbacks
        bpuc.CallbackManager.add_callback_to_attribute(target_class=type(bpnc_area), target_instance=bpnc_area, attribute_name='_update_count', callback=self.on_update_count_update)
        bpuc.CallbackManager.add_callback_to_attribute(target_class=type(briann), target_instance=briann, attribute_name='_current_simulation_time', callback=self.on_current_simulation_time_update)
        bpuc.CallbackManager.add_callback_to_method(target_instance=bpnc_area, method_name='reset', callback= lambda caller, selfy=self: selfy.on_current_simulation_time_update(obj=None, name=None, value=None)) # Reset color when area is reset
        
        # Create button
        self._button = customtkinter.CTkButton(canvas, 
                                         text = bpnc_area.index, 
                                         fg_color='lightgray', 
                                         border_color="darkgray",#"#325882",# "#14375e", 
                                         border_width=0.05*size*DPI, 
                                         text_color='black', 
                                         anchor = tk.CENTER, 
                                         width=size*DPI, # Dynamcially adjusts width based on text length
                                         height=size*DPI, 
                                         corner_radius=0.5*size*DPI)
        
        # Add right-click functionality 
        self._button.bind("<Button-2>", self._on_right_click)
        self._button.bind("<Button-3>", self._on_right_click)

        # Add state visualizers array
        self._state_visualizers = []

        # Call super
        super().__init__(canvas=canvas, widget=self._button, x=x, y=y)

    @property
    def size(self) -> float:
        """:return: The size of this area in inches.
        :rtype: float"""
        return self._size
    
    @property
    def state_visualizers(self) -> List["AreaStateVisualizer"]:
        """:return: The list of state visualizers that are currently attached to this area.
        :rtype: List[:py:class:`.AreaStateVisualizer`]"""
        return self._state_visualizers

    def _on_right_click(self, event: tk.Event) -> None:
        """Displays a pop-up window that allow to choose a :py:class:`.AreaStateVisualizer`.
        
        :param event: The event that triggered the function call.
        :type event: :py:class:`tkinter.Event`
        :rtype: None
        """
        
        # Create a popup
        popup = tk.Toplevel()
        popup.geometry(f"{(int)(2*DPI)}x{(int)(2*DPI)}+{self._canvas.winfo_rootx()+self._button.winfo_x()}+{self._canvas.winfo_rooty()+self._button.winfo_y()}")
        popup.overrideredirect(True) # Prevent window decorations
        
        # Add widgets to popup
        customtkinter.CTkLabel(popup, text="Add Visualizer", anchor=tk.CENTER).pack(expand=True, fill="x", padx=10, pady=10)
        
        option_values = []
        for visualizer in [StateVisualizerLineChart, StateVisualizerHeatMap]:
            if visualizer.is_compatible_with_state(input_shape=self._bpnc_area.output_time_frame_accumulator._time_frame.state.shape):
                option_values.append(visualizer.NAME_TAG)

        option_menu = customtkinter.CTkOptionMenu(popup, values=option_values)
        option_menu.pack(expand=True, fill="x", padx=10, pady=10)
        customtkinter.CTkButton(popup, text="Add", 
                                command=lambda option_menu=option_menu, popup=popup, area=self: area._add_state_visualizer(option=option_menu.get(), popup=popup)
                                ).pack(expand=True, fill="x", padx=10, pady=10)
        
        # Display popup on top of everything else
        popup.lift()
             
    def _add_state_visualizer(self, option: str, popup: customtkinter.CTkToplevel) -> None:
        """Maps the user's selection for the preferred :py:class:`.AreaStateVisualizer` to the actual visualizer to be displayed on screen.
        
        :param option: The name of the chosen visualizer. Options are ["Line Chart"].
        :type option: str
        :param popup: The pop-up that allowed the users to make a choice. This pop-up will be deleted by the current method.
        :type popup: :py:class:`customtkinter.CTkTopLevel`
        :rtype: None
        """

        # Map selection option to visualizer
        if option == "Line Chart":
            self._state_visualizers.append(StateVisualizerLineChart(area=self._bpnc_area, canvas=self._canvas, current_simulation_time=self._briann.current_simulation_time, initial_x=self.x, initial_y=self.y, width=3, height=2))
        if option == "Heatmap":
            self._state_visualizers.append(StateVisualizerHeatMap(area=self._bpnc_area, canvas=self._canvas, current_simulation_time=self._briann.current_simulation_time, initial_x=self.x, initial_y=self.y, width=3, height=2))

        # Destroy the popup
        popup.destroy()
    
    def on_current_simulation_time_update(self, obj, name, value) -> None:
        
        # Display self as inactive
        self._button.configure(fg_color='lightgray', text_color='black')

    def on_update_count_update(self, obj, name, value) -> None:
        
        # Display self as active
        self._button.configure(fg_color='orange', text_color='white')

class Connection(DraggableWidget.LocationSubscriber):
    """A visual representation of a :py:class:`~src.briann.python.components.Connection`. The connection is automatically redraws itself when the area is repositioned.
    
    :param from_area: The area from which the connection starts.
    :type from_area: :py:class:`.Area`
    :param to_area: The area at which the connection ends.
    :type to_area: :py:class:`.Area`
    :canvas: The canvas on which the connection shall be drawn.
    :type canvas: :py:class:`.Canvas`
    :param width: The thickness of the line representing this connection on screen.
    :type width: int, optional, default to 2.
    :param bend_by: The extend by which the connection should be bent (in inches) relative to the mid-point between the `from_area` and the `to_area`.
    :type bend_by: float, optional, defaults to 0.0.
    :return: An instance of this class.
    :rtype: :py:class:`.Connection`  
    """

    def __init__(self, from_area: Area, to_area: Area, canvas: Canvas, width: int = 2, bend_by: float = 0.0) -> "Connection":
        
        # Set properties
        self._canvas = canvas
        self._from_area = from_area
        self._to_area = to_area
        self._width = width
        self._bend_by = bend_by
        
        # Draw connection
        self.draw()

        # Add self as location subscriber to areas
        from_area.add_location_subscriber(subscriber=self)
        to_area.add_location_subscriber(subscriber=self)

    def draw(self) -> None:
        """Draws self on the canvas as two segments. The first segment goes from the from-area to the mid-point with an arrow-head. The second segment goes from the mid-point to the to-area.
        
        :rtype: None"""

        # Get start end points from areas
        x0, y0 = self._from_area.x, self._from_area.y # Starting point
        x1, y1 = self._to_area.x, self._to_area.y # Endpoint
        
        # Self loop
        if x0 == x1 and y0 == y1:
            
            # Bounding box for circle
            top_left = (x0,y0)
            a = self._from_area.size # Step size for x and y from top left to bottom right
            bottom_right = (x0 + a, y0 - a)
            b = np.sqrt((a/2.0)**2+(a/2.0)**2) # length of diagonal from center of box to its corner
            tmp = np.cos(np.radians(45)) * (b + (a/2.0)) # Projection onto one of the sides of the bounding box
            arrow_position = (x0+tmp, y0-tmp)

            # Draw on screen space
            top_left = Canvas.cartesian_to_canvas(x=top_left[0], y=top_left[1])
            bottom_right = Canvas.cartesian_to_canvas(x=bottom_right[0], y=bottom_right[1])
            arrow_position = Canvas.cartesian_to_canvas(x=arrow_position[0], y=arrow_position[1])
            self._first_segment = self._canvas.create_arc(*top_left, *bottom_right, width=self._width, start=0, extent=359, style='arc')
            self._second_segment = self._canvas.create_line(arrow_position[0]-1, arrow_position[1]+1, *arrow_position, arrow='last')
            return

        # Straight line
        if self._bend_by == 0:
            # Draw on screen space
            x0, y0 = Canvas.cartesian_to_canvas(x=x0, y=y0)
            x1, y1 = Canvas.cartesian_to_canvas(x=x1, y=y1)
            xm, ym = (x0+x1)/2, (y0+y1)/2
            self._first_segment = self._canvas.create_line(x0,y0, xm,ym, width=self._width, arrow='last'); 
            self._second_segment = self._canvas.create_line(xm,ym, x1,y1, width=self._width); 
            
            return
        
        else:
            # Arc (tilted and translated ellipse segment)
            height = 2*self._bend_by
            width = np.sqrt((x1-x0)**2+(y1-y0)**2)
            
            y = lambda x: np.sqrt(((height/2)**2)*(1 - (x**2)/((width/2)**2))) # Equation of an ellipse centered at (0,0) with radii width/2 and height/2
            
            n = 20
            arc_points = np.linspace(start=-width/2, stop=width/2, num=n)
            arc_points = np.array([[x, y(x)] for x in arc_points]).T
            angle = np.arctan2(y1-y0, x1-x0) * 180 / np.pi # Angle between the two areas in degrees
            R = np.array([[np.cos(np.radians(angle)), -np.sin(np.radians(angle))], [np.sin(np.radians(angle)), np.cos(np.radians(angle))]]) # Rotation matrix
            arc_points = np.dot(R,arc_points) + np.array([[(x0+x1)/2], [(y0+y1)/2]]) # Apply rotation and translation
            
            # Draw on screen space
            arc_points = [Canvas.cartesian_to_canvas(x=arc_points[0,i], y=arc_points[1,i]) for i in range(arc_points.shape[1])]
            self._first_segment = self._canvas.create_line(*arc_points[:len(arc_points)//2+1], width=self._width, smooth=True, arrow='last')
            self._second_segment = self._canvas.create_line(*arc_points[len(arc_points)//2:], width=self._width, smooth=True)
            
    def on_location_update(self, draggable_widget: DraggableWidget) -> None:
        
        # Remove old segments
        self._canvas.delete(self._first_segment)
        self._canvas.delete(self._second_segment)
        
        # Draw new segments
        self.draw()

class NetworkVisualizer():
    """Visualizes the network structure of a given `briann` instance in a rectangle whose center is at (`initial_x`,`initial_y`) with provided `width` and `height`.
        The visualizer will be drawn on top of all previously drawn elements on the `canvas` and it can be dragged around with the mouse.

        :param briann: The BrIANN instance to visualize.
        :type briann: bpnc.BrIANN
        :param canvas: The canvas to draw the visualizer on.
        :type canvas: Canvas
        :param x: The x-coordinate of the visualizer's center in Cartesian space (inches).
        :type x: float
        :param y: The y-coordinate of the visualizer's center in Cartesian space (inches).
        :type y: float
        :param width: The width of the visualizer in inches.
        :type width: float
        :param height: The height of the visualizer in inches.
        :type height: float
        :return: The visualizer that was created.
        :rtype: :py:class:`.NetworkVisualizer`"""

    def __init__(self, briann: bpnc.BrIANN, canvas: Canvas, initial_x: float, initial_y: float, width: float, height: float, area_size: float) -> "NetworkVisualizer":
        
        # Set properties
        self.canvas = canvas
        self._briann = briann
        
        # Override the drag functions from super to extend it to all child widgets

        # Compute positions
        G = self.briann.get_topology()
        area_to_position = nx.shell_layout(G=G, center=(0,0))

        # Convert from G space to canvas space
        x_min, x_max = min([pos[0] for pos in area_to_position.values()]), max([pos[0] for pos in area_to_position.values()]) # G space
        y_min, y_max = min([pos[1] for pos in area_to_position.values()]), max([pos[1] for pos in area_to_position.values()]) # G space
        x_range, y_range = x_max-x_min, y_max-y_min # G space
        for area, position in area_to_position.items():
            x, y = position[0] / (0.5*x_range), position[1] / (0.5*y_range) # Ensure width and height of whole network is between -1 and 1
            x, y = x*0.5*(width-area_size), y*0.5*(height-area_size) # Scale to visualizer size
            area_to_position[area] = (x+initial_x, y+initial_y)

        # Create area drawables
        self.area_to_drawable = {}
        for area in self.briann.areas:
            x, y = area_to_position[area][0], area_to_position[area][1]
            self.area_to_drawable[area] = Area(bpnc_area=area, briann=self._briann, canvas=self.canvas, x=x, y=y, size=area_size)
            
        # Draw the edges
        width = 0.05*area_size
        curved_edges = [edge for edge in G.edges() if reversed(edge) in G.edges()]
        self.edge_to_drawable = {}
        for (u,v) in G.edges():
            # Convert index to area
            u = self.briann.get_area_at_index(index=u.index)
            v = self.briann.get_area_at_index(index=v.index)
            bend_by = 0.25 if (u,v) in curved_edges else 0.0
            self.edge_to_drawable[u.index, v.index] = Connection(from_area=self.area_to_drawable[u], to_area=self.area_to_drawable[v], canvas=self.canvas, width=width, bend_by=bend_by)
          
    @property
    def briann(self) -> bpnc.BrIANN:
        """
        :return: The BrIANN instance that is visualized by this visualizer.
        :rtype: bpnc.BrIANN"""
        return self._briann

class AreaStateVisualizer(DraggableWidget):
    """Superclass for a set of classes that create 2D visualizations of a :py:meth:`.TimeFrame.state` on a 1x1 unit square"""

    def __init__(self, area: bpnc.Area, canvas: Canvas, current_simulation_time: float, initial_x: float, initial_y: float, width: float, height: float):
    
        # Set proeprties
        self.bpnc = area

        # Create Figure
        self.figure = plt.figure(figsize=(width, height), dpi=DPI)
        
        self.figure.set_tight_layout(True)
        widget = FigureCanvasTkAgg(plt.gcf()).get_tk_widget()
        
        # Call super
        super().__init__(canvas=canvas, widget=widget, x=initial_x, y=initial_y)

        # Subscribe to area
        bpuc.CallbackManager.add_callback_to_method(target_instance=area, method_name='forward', callback=self.on_forward_call)
        bpuc.CallbackManager.add_callback_to_method(target_instance=area, method_name='reset', callback=self.on_reset_call)
        
        # Initial draw
        self._update_plot(time_frame = area.output_time_frame_accumulator.time_frame(current_time=current_simulation_time)) # Draw current time frame
        
    def on_forward_call(self, caller) -> None:
        time_frame = self.bpnc.output_time_frame_accumulator._time_frame
        self._update_plot(time_frame=time_frame)

    def on_reset_call(self, caller) -> None:
        # Reset data of self
        self.ts = np.empty(shape=[0,0])
        self.ys = np.empty(shape=[0,0])
        
        # Redraw
        time_frame = self.bpnc.output_time_frame_accumulator._time_frame
        self._update_plot(time_frame=time_frame)

    def _update_plot(self, time_frame: bpnc.TimeFrame = None) -> None:
        plt.figure(self.figure.number)
        
class StateVisualizerLineChart(AreaStateVisualizer):
    """Plots a series of :py:clas:`~briann.python.network.core.TimeFrame`'s in a line-plot where the x-axis represents time, the y-axis represents deflection and there will be one line per channel.
    This class assumes that the time-series is streamed to py:meth:`~briann.python.GUI.StateVisualizerLineChart._update_plot_` one time-frame at a time.
    Each such time-frame is expected to be of shape (batch-size, channel-count). """

    NAME_TAG = "Line Chart"

    def __init__(self, area: bpnc.Area, canvas: Canvas, current_simulation_time: float, initial_x: float, initial_y: float, width: float, height: float) -> "StateVisualizerLineChart":
        
        # FIRST Set Attributes (needed for initial plot)
        self.ts = np.empty(shape=[0,0])
        self.ys = np.empty(shape=[0,0])
        
        # Then Call super
        super().__init__(area=area, canvas=canvas, current_simulation_time=current_simulation_time, initial_x=initial_x, initial_y=initial_y, width=width, height=height)
        

    @classmethod
    def is_compatible_with_state(cls, input_shape: List[int]) -> bool:
        """Checks whether the given input shape is compatible with this visualizer.
        
        :param input_shape: The time-frame's state's shape to be checked.
        :type input_shape: List[int]
        :return: True if the shae is compatible with this visualizer, False otherwise.
        :rtype: bool"""
        return len(input_shape) == 2 # (batch_size, features)

    def _update_plot(self, time_frame: bpnc.TimeFrame = None) -> None:
        
        # Call super
        super()._update_plot(time_frame=time_frame)

        # Update plot
        plt.clf()
                 
        if len(self.ts) == 0: self.ts = np.repeat([time_frame.time_point], time_frame.state.shape[-1])[np.newaxis,:]
        else: self.ts = np.concatenate([self.ts, np.repeat([time_frame.time_point], time_frame.state.shape[-1])[np.newaxis,:]], axis=0)
    
        if len(self.ys) == 0: self.ys = time_frame.state.cpu().detach().numpy()[0,:][np.newaxis,:]
        else: self.ys = np.concatenate([self.ys, time_frame.state.cpu().detach().numpy()[0,:][np.newaxis,:]], axis=0)  
        
        plt.plot(self.ts, self.ys)
        for channel in range(self.ys.shape[1]): plt.scatter(self.ts[:,channel], self.ys[:,channel])
        plt.xlabel("Time (s)")
        plt.ylabel("State")
        plt.title(f"Area {self.bpnc.index}")
        plt.draw()

class StateVisualizerHeatMap(AreaStateVisualizer):
    """Plots a series of :py:clas:`~briann.python.network.core.TimeFrame`'s in an evolving heatmap where the x-axis represents width, the y-axis represents height and the heatmap represents values.
    It assumes that the input state has a shape of (batch-size, channels, height, width) where channels is 1 or 3 (grayscale or RGB) or any other number of channels which will then be averaged across.
    An alternative accepted shape is (batch-size, height, width) for single-channel data.
    The visualizer always plots only the first instance of the batch (index 0)."""

    NAME_TAG = "Heatmap"

    def __init__(self, area: bpnc.Area, canvas: Canvas, current_simulation_time: float, initial_x: np.ndarray, initial_y: float, width: float, height: float) -> "StateVisualizerLineChart":
        
        super().__init__(area=area, canvas=canvas, current_simulation_time=current_simulation_time, initial_x=initial_x, initial_y=initial_y, width=width, height=height)
        
    
    @classmethod
    def is_compatible_with_state(cls, input_shape: List[int]) -> bool:
        """Checks whether the given input shape is compatible with this visualizer.
        
        :param input_shape: The time-frame's state's shape to be checked.
        :type input_shape: List[int]
        :return: True if the shae is compatible with this visualizer, False otherwise.
        :rtype: bool"""
        return len(input_shape) == 3 or (len(input_shape) == 4) # (batch_size, width, height) or (batch_size, channels, width, height)


    def _update_plot(self, time_frame: bpnc.TimeFrame = None) -> None:
        
        # Call super
        super()._update_plot(time_frame=time_frame)

        # Update plot
        plt.clf()
        
        if len(time_frame.state.shape) == 3: # (batch_size, height, width)
            plt.imshow(time_frame.state[0].cpu().detach().numpy().T, aspect='auto') # We take the first instance [0]
        else:
            if len(time_frame.state.shape) == 4: # (batch_size, channel-count, height, width)
                state = time_frame.state[0].cpu().detach().numpy()  # We take the first instance [0]
                if state.shape[1] != 1 and state.shape[0] != 3: # Not just 1 or 3 channels, average across channels
                    state = np.mean(state, axis=0, keepdims=True) # Shape now (1 channel, height, width)
                
                state = np.moveaxis(state, 0, -1) # Move channel axis to the end, shape now (height, width, channel-count)
                state = np.moveaxis(state, 0, 1) # transpose width and height, shape now (width, height, channel-count)
                plt.imshow(state, aspect='auto')
        
        plt.title(f"Area {self.bpnc.index}")
        plt.draw()
    


# Copyright 2017 The TensorFlow Authors All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Compute input examples for VGGish from audio waveform."""

import numpy as np
import sys, os
sys.path.append(os.path.abspath(""))
from briann.utilities import mel_features, vggish_params
import torch

def waveform_to_examples(data):
  """Converts audio waveform into an array of examples for VGGish.

  Args:
    data: np.array of either one dimension (mono) or two dimensions
      (multi-channel, with the outer dimension representing channels).
      Each sample is expected to be sampled at 16kHz and to lie in the range [-1.0, +1.0],

  Returns:
    3-D np.array of shape [num_examples, num_frames, num_bands] which represents
    a sequence of examples, each of which contains a patch of log mel
    spectrogram, covering num_frames frames of audio and num_bands mel frequency
    bands, where the frame length is vggish_params.STFT_HOP_LENGTH_SECONDS.
  """
  # Convert to mono.
  if len(data.shape) > 1:
    data = np.mean(data, axis=1)
  # Resample to the rate assumed by VGGish.
  
  # Compute log mel spectrogram features.
  log_mel = mel_features.log_mel_spectrogram(
      data,
      audio_sample_rate=vggish_params.SAMPLE_RATE,
      log_offset=vggish_params.LOG_OFFSET,
      window_length_secs=vggish_params.STFT_WINDOW_LENGTH_SECONDS,
      hop_length_secs=vggish_params.STFT_HOP_LENGTH_SECONDS,
      num_mel_bins=vggish_params.NUM_MEL_BINS,
      lower_edge_hertz=vggish_params.MEL_MIN_HZ,
      upper_edge_hertz=vggish_params.MEL_MAX_HZ)

  # Frame features into examples.
  features_sample_rate = 1.0 / vggish_params.STFT_HOP_LENGTH_SECONDS
  example_window_length = int(round(
      vggish_params.EXAMPLE_WINDOW_SECONDS * features_sample_rate))
  example_hop_length = int(round(
      vggish_params.EXAMPLE_HOP_SECONDS * features_sample_rate))
  log_mel_examples = mel_features.frame(
      log_mel,
      window_length=example_window_length,
      hop_length=example_hop_length)
  
  log_mel_examples = torch.tensor(
            log_mel_examples, requires_grad=True)[:, None, :, :].float()

  return log_mel_examples


from typing import List, Iterator
from torch.utils.data import Sampler
import numpy as np

class StatelessRandom():
        
    def linear_congruential_generator(multiplier: int, modulo: int, seed: int, count: int) -> np.ndarray:
        
        # Compute array of random numbers
        result = np.zeros(shape=count)
        tmp = (seed*multiplier+1) % modulo
        result[0] = tmp/modulo
        for i in range(1, count):
            tmp = (tmp*multiplier+1) % modulo
            result[i] = tmp / modulo

        # Output
        return result

    def pseudo_uniform(lower_bound: float, upper_bound: float, seed: int, count: int) -> np.ndarray:

        # Get random numbers in range 0 to 1
        x = StatelessRandom.linear_congruential_generator(multiplier=3467541, modulo=(2**31)-1, seed=seed, count=count)

        # Scale
        x = lower_bound+(upper_bound-lower_bound)*x

        # Output
        return x

    def shuffle(indices: List[int], seed: int) -> List[int]:

        random_numbers = StatelessRandom.pseudo_uniform(lower_bound=0, upper_bound=len(indices), seed=seed, count=len(indices))

        # Iterate indices
        for i in range(len(indices)):
            # Find candidate to swop positions with
            j = (int)(random_numbers[i])
            
            # Swop
            tmp = indices[j]
            indices[j] = indices[i]
            indices[i] = tmp

class IndexSampler(Sampler[int]):

    def __init__(self, instance_count: int, seed: int = None) -> None:
        self._instance_count = instance_count
        self._seed = seed
        self._indices = list(range(instance_count))
        StatelessRandom.shuffle(indices=self._indices, seed=seed)

    @property
    def instance_count(self) -> int:
        return self._instance_count
    
    @property
    def seed(self) -> int:
        return self._seed

    def __len__(self) -> int:
        return self.instance_count
    
    def __iter__(self) -> Iterator[int]:
        for i in range(self.instance_count):
            yield self._indices[i]

if __name__ == "__main__":
    sampler = IndexSampler(instance_count=10, seed=5)
    for i in sampler:
        print(i)

    for i in sampler:
        print(i)

        
    import tkinter as tk
    import turtle

    def run_turtles(*args):
        for t, d in args:
            t.circle(200, d)
        root.after_idle(run_turtles, *args)

    def scroll_start(event):
        screen.scan_mark(event.x, event.y)

    def scroll_move(event):
        screen.scan_dragto(event.x, event.y, gain=1)

    root = tk.Tk()
    root.geometry("700x700")
    root.withdraw()

    frame = tk.Frame(bg='black')
    frame.pack(fill='both', expand=True)
    tk.Label(frame, text=u'Hello', bg='grey', fg='white').pack(fill='x')

    screen = turtle.ScrolledCanvas(frame)
    screen.pack(fill="both", expand=True)


    turtle1 = turtle.RawTurtle(screen)
    turtle2 = turtle.RawTurtle(screen)

    screen.bind("<ButtonPress-1>", scroll_start)
    screen.bind("<B1-Motion>", scroll_move)

    turtle1.ht(); turtle1.pu()
    turtle1.left(90); turtle1.fd(200); turtle1.lt(90)
    turtle1.st(); turtle1.pd()

    turtle2.ht(); turtle2.pu()
    turtle2.fd(200); turtle2.lt(90)
    turtle2.st(); turtle2.pd()

    root.deiconify()

    run_turtles((turtle1, 3), (turtle2, 4))

    root.mainloop()

        


# Copyright 2017 The TensorFlow Authors All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Defines routines to compute mel spectrogram features from audio waveform."""

import numpy as np


def frame(data, window_length, hop_length):
  """Convert array into a sequence of successive possibly overlapping frames.

  An n-dimensional array of shape (num_samples, ...) is converted into an
  (n+1)-D array of shape (num_frames, window_length, ...), where each frame
  starts hop_length points after the preceding one.

  This is accomplished using stride_tricks, so the original data is not
  copied.  However, there is no zero-padding, so any incomplete frames at the
  end are not included.

  Args:
    data: np.array of dimension N >= 1.
    window_length: Number of samples in each frame.
    hop_length: Advance (in samples) between each window.

  Returns:
    (N+1)-D np.array with as many rows as there are complete frames that can be
    extracted.
  """
  num_samples = data.shape[0]
  num_frames = 1 + int(np.floor((num_samples - window_length) / hop_length))
  shape = (num_frames, window_length) + data.shape[1:]
  strides = (data.strides[0] * hop_length,) + data.strides
  return np.lib.stride_tricks.as_strided(data, shape=shape, strides=strides)


def periodic_hann(window_length):
  """Calculate a "periodic" Hann window.

  The classic Hann window is defined as a raised cosine that starts and
  ends on zero, and where every value appears twice, except the middle
  point for an odd-length window.  Matlab calls this a "symmetric" window
  and np.hanning() returns it.  However, for Fourier analysis, this
  actually represents just over one cycle of a period N-1 cosine, and
  thus is not compactly expressed on a length-N Fourier basis.  Instead,
  it's better to use a raised cosine that ends just before the final
  zero value - i.e. a complete cycle of a period-N cosine.  Matlab
  calls this a "periodic" window. This routine calculates it.

  Args:
    window_length: The number of points in the returned window.

  Returns:
    A 1D np.array containing the periodic hann window.
  """
  return 0.5 - (0.5 * np.cos(2 * np.pi / window_length *
                             np.arange(window_length)))


def stft_magnitude(signal, fft_length,
                   hop_length=None,
                   window_length=None):
  """Calculate the short-time Fourier transform magnitude.

  Args:
    signal: 1D np.array of the input time-domain signal.
    fft_length: Size of the FFT to apply.
    hop_length: Advance (in samples) between each frame passed to FFT.
    window_length: Length of each block of samples to pass to FFT.

  Returns:
    2D np.array where each row contains the magnitudes of the fft_length/2+1
    unique values of the FFT for the corresponding frame of input samples.
  """
  frames = frame(signal, window_length, hop_length)
  # Apply frame window to each frame. We use a periodic Hann (cosine of period
  # window_length) instead of the symmetric Hann of np.hanning (period
  # window_length-1).
  window = periodic_hann(window_length)
  windowed_frames = frames * window
  return np.abs(np.fft.rfft(windowed_frames, int(fft_length)))


# Mel spectrum constants and functions.
_MEL_BREAK_FREQUENCY_HERTZ = 700.0
_MEL_HIGH_FREQUENCY_Q = 1127.0


def hertz_to_mel(frequencies_hertz):
  """Convert frequencies to mel scale using HTK formula.

  Args:
    frequencies_hertz: Scalar or np.array of frequencies in hertz.

  Returns:
    Object of same size as frequencies_hertz containing corresponding values
    on the mel scale.
  """
  return _MEL_HIGH_FREQUENCY_Q * np.log(
      1.0 + (frequencies_hertz / _MEL_BREAK_FREQUENCY_HERTZ))


def spectrogram_to_mel_matrix(num_mel_bins=20,
                              num_spectrogram_bins=129,
                              audio_sample_rate=8000,
                              lower_edge_hertz=125.0,
                              upper_edge_hertz=3800.0):
  """Return a matrix that can post-multiply spectrogram rows to make mel.

  Returns a np.array matrix A that can be used to post-multiply a matrix S of
  spectrogram values (STFT magnitudes) arranged as frames x bins to generate a
  "mel spectrogram" M of frames x num_mel_bins.  M = S A.

  The classic HTK algorithm exploits the complementarity of adjacent mel bands
  to multiply each FFT bin by only one mel weight, then add it, with positive
  and negative signs, to the two adjacent mel bands to which that bin
  contributes.  Here, by expressing this operation as a matrix multiply, we go
  from num_fft multiplies per frame (plus around 2*num_fft adds) to around
  num_fft^2 multiplies and adds.  However, because these are all presumably
  accomplished in a single call to np.dot(), it's not clear which approach is
  faster in Python.  The matrix multiplication has the attraction of being more
  general and flexible, and much easier to read.

  Args:
    num_mel_bins: How many bands in the resulting mel spectrum.  This is
      the number of columns in the output matrix.
    num_spectrogram_bins: How many bins there are in the source spectrogram
      data, which is understood to be fft_size/2 + 1, i.e. the spectrogram
      only contains the nonredundant FFT bins.
    audio_sample_rate: Samples per second of the audio at the input to the
      spectrogram. We need this to figure out the actual frequencies for
      each spectrogram bin, which dictates how they are mapped into mel.
    lower_edge_hertz: Lower bound on the frequencies to be included in the mel
      spectrum.  This corresponds to the lower edge of the lowest triangular
      band.
    upper_edge_hertz: The desired top edge of the highest frequency band.

  Returns:
    An np.array with shape (num_spectrogram_bins, num_mel_bins).

  Raises:
    ValueError: if frequency edges are incorrectly ordered or out of range.
  """
  nyquist_hertz = audio_sample_rate / 2.
  if lower_edge_hertz < 0.0:
    raise ValueError("lower_edge_hertz %.1f must be >= 0" % lower_edge_hertz)
  if lower_edge_hertz >= upper_edge_hertz:
    raise ValueError("lower_edge_hertz %.1f >= upper_edge_hertz %.1f" %
                     (lower_edge_hertz, upper_edge_hertz))
  if upper_edge_hertz > nyquist_hertz:
    raise ValueError("upper_edge_hertz %.1f is greater than Nyquist %.1f" %
                     (upper_edge_hertz, nyquist_hertz))
  spectrogram_bins_hertz = np.linspace(0.0, nyquist_hertz, num_spectrogram_bins)
  spectrogram_bins_mel = hertz_to_mel(spectrogram_bins_hertz)
  # The i'th mel band (starting from i=1) has center frequency
  # band_edges_mel[i], lower edge band_edges_mel[i-1], and higher edge
  # band_edges_mel[i+1].  Thus, we need num_mel_bins + 2 values in
  # the band_edges_mel arrays.
  band_edges_mel = np.linspace(hertz_to_mel(lower_edge_hertz),
                               hertz_to_mel(upper_edge_hertz), num_mel_bins + 2)
  # Matrix to post-multiply feature arrays whose rows are num_spectrogram_bins
  # of spectrogram values.
  mel_weights_matrix = np.empty((num_spectrogram_bins, num_mel_bins))
  for i in range(num_mel_bins):
    lower_edge_mel, center_mel, upper_edge_mel = band_edges_mel[i:i + 3]
    # Calculate lower and upper slopes for every spectrogram bin.
    # Line segments are linear in the *mel* domain, not hertz.
    lower_slope = ((spectrogram_bins_mel - lower_edge_mel) /
                   (center_mel - lower_edge_mel))
    upper_slope = ((upper_edge_mel - spectrogram_bins_mel) /
                   (upper_edge_mel - center_mel))
    # .. then intersect them with each other and zero.
    mel_weights_matrix[:, i] = np.maximum(0.0, np.minimum(lower_slope,
                                                          upper_slope))
  # HTK excludes the spectrogram DC bin; make sure it always gets a zero
  # coefficient.
  mel_weights_matrix[0, :] = 0.0
  return mel_weights_matrix


def log_mel_spectrogram(data,
                        audio_sample_rate=8000,
                        log_offset=0.0,
                        window_length_secs=0.025,
                        hop_length_secs=0.010,
                        **kwargs):
  """Convert waveform to a log magnitude mel-frequency spectrogram.

  Args:
    data: 1D np.array of waveform data.
    audio_sample_rate: The sampling rate of data.
    log_offset: Add this to values when taking log to avoid -Infs.
    window_length_secs: Duration of each window to analyze.
    hop_length_secs: Advance between successive analysis windows.
    **kwargs: Additional arguments to pass to spectrogram_to_mel_matrix.

  Returns:
    2D np.array of (num_frames, num_mel_bins) consisting of log mel filterbank
    magnitudes for successive frames.
  """
  window_length_samples = int(round(audio_sample_rate * window_length_secs))
  hop_length_samples = int(round(audio_sample_rate * hop_length_secs))
  fft_length = 2 ** int(np.ceil(np.log(window_length_samples) / np.log(2.0)))
  spectrogram = stft_magnitude(
      data,
      fft_length=fft_length,
      hop_length=hop_length_samples,
      window_length=window_length_samples)
  mel_spectrogram = np.dot(spectrogram, spectrogram_to_mel_matrix(
      num_spectrogram_bins=spectrogram.shape[1],
      audio_sample_rate=audio_sample_rate, **kwargs))
  return np.log(mel_spectrogram + log_offset)
